structure Continuations :> NUBOX =
struct
    structure Support = SupportList

    structure Types = 
	ObjectTypes (structure Support = Support)

    structure Absyn = 
	AbstractSyntax (structure Support = SupportList
			structure Types = Types)
	
    structure Parse =
	Parsing (structure Absyn = Absyn)
	
    structure Ctx = 
	Context (structure Absyn = Absyn)
	
    structure Typecheck = 
	Typechecking (structure Absyn = Absyn
		      structure Ctx = Ctx
		      val lang = Language.CONTS) 
	
	
    exception RuntimeError

    type var = Var.var
    open Absyn

    fun isValue (v : term) : bool =
	(case v of 
	     Empty => true
	   | Nat n => true
	   | True => true
	   | False => true
	   | Nil => true
	   | Cons (v1, v2) => 
		 (isValue v1) andalso (isValue v2)
	   | Fn (x, e) => true
	   | Fix (f, x, e) => true
	   | Pair (v1, v2) => 
		 (isValue v1) andalso (isValue v2)
	   | Box e => true
	   | Nu (x, e) => true
	   | _ => false)


    datatype cont = Return | Cont of term -> term
    withtype stack = (var * cont) list


    val internal = Var.fromString_unique "internal"

    val emptystack = [(internal, Return)]

    fun push (s : stack, f : term -> term) : stack =
	(case s of 
	     nil => raise RuntimeError
	   | ((x, Return)::ss) => (x, Cont f)::ss
	   | ((x, Cont cc)::ss) => 
		 let val cc' : term->term = 
		     fn c => cc (f c)
		 in
		     (x, Cont cc')::ss
		 end)
	     
    fun pop (s : stack) : var * cont * stack =
	(case s of
	     nil => raise RuntimeError
	   | ((x, cont)::ss) => (x, cont, ss))


    fun apply (e : term, s : stack) : term =
	(case s of
	     nil => e
	   | ((l, cont)::ss) => 
		 (case cont of
		      Return => apply (e, ss)
		    | Cont cc => evalTerm (cc e, (l, Return)::ss)))
		    

    (***
     each evaluation function takes a static environment 
     providing values and/or expressions for variables 
     ***)

     
    and evalTerm (e : term, stack : stack) : term =
	(case e of
	     Empty => apply (e, stack)
	   | Print (e') => 
		 if (isValue e') then 
		     let val _ = print ((termToString e')^"\n")
		     in
			 apply (Empty, stack)
		     end
		 else
		     evalTerm (e', push (stack, fn x => Print x))
	   | Nat x => apply (e, stack)
	   | True =>  apply (e, stack)
	   | False => apply (e, stack)
	   | Nil => apply (e, stack)
	   | Cons (e1, e2) => 
		 (if (isValue e1) then
		      if (isValue e2) then apply (Cons (e1, e2), stack)
		      else
			  evalTerm (e2, push (stack, fn y => Cons (e1, y)))
		  else
		      evalTerm (e1, push (stack, fn y => Cons (y, e2))))
		      
	   | Car (e1) => 
		 (if (isValue e1) then
		      let val Cons (hd, tl) = e1
		      in
			  apply (hd, stack)
		      end
		  else
		      evalTerm (e1, push (stack, fn y => Car y)))

	   | Cdr (e1) => 
		 (if (isValue e1) then
		      let val Cons (hd, tl) = e1
		      in
			  apply (tl, stack)
		      end
		  else
		      evalTerm (e1, push (stack, fn y => Cdr y)))


	   | BinOp (p, e1, e2) => 
		 (if (isValue e1) then
		      if (isValue e2) then
			  case (p, e1, e2) of
			      (MINUS, Nat n1, Nat n2) => apply (Nat (n1 - n2), stack)
			    | (PLUS, Nat n1, Nat n2) => apply (Nat (n1 + n2), stack)
			    | (MULT, Nat n1, Nat n2) => apply (Nat (n1 * n2), stack)
			    | (EQL, e1, e2) => 
				  if termEq (e1, e2) then apply (True, stack)
				  else apply (False, stack)
		      else
			  evalTerm (e2, push (stack, fn y => BinOp (p, e1, y)))
		  else
		      evalTerm (e1, push (stack, fn x => BinOp (p, x, e2))))

	   | If (True, e1, e2) => 
		  evalTerm (e1, stack)

	   | If (False, e1, e2) => 
		  evalTerm (e2, stack)

	   | If (c, e1, e2) => 
		  evalTerm (c, push (stack, fn c => If (c, e1, e2)))

	   | Fn (x, e') => apply (e, stack)

	   | Fix (f, x, e') => apply (e, stack)

	   | App (e1 as Fn (x, e3), e2) => 
		  (if (isValue e2) then 
		       let val s = Var.VarMap.insert (Var.VarMap.empty, x, e2)
		       in 
			   evalTerm (substTerm (s, e3), stack)
		       end
		   else
		       evalTerm (e2, push (stack, fn y => App (e1, y))))
		       
	   | App (e1 as (Fix (f, x, e3)), e2) => 
		  (if (isValue e2) then 
		       let val s1 = Var.VarMap.insert (Var.VarMap.empty, f, e1)
			   val s2 = Var.VarMap.insert (s1, x, e2)
			   val e3' = substTerm (s2, e3)
		       in
			   evalTerm (e3', stack)
		       end
		   else
		       evalTerm (e2, push (stack, fn y => App (e1, y))))
		       
	   | App (e1, e2) => 
		 evalTerm (e1, push (stack, fn y => App (y, e2)))

	   | Pair (e1, e2) => 
		 (if (isValue e1) then 
		      if (isValue e2) then apply (e, stack)
		      else
			  evalTerm (e2, push (stack, fn y => Pair (e1, y)))
		  else
		      evalTerm (e1, push (stack, fn y => Pair (y, e2))))

	   | Fst (e') => 
		 if (isValue e') then
		     let val Pair (v1, v2) = e'
		     in
			 apply (v1, stack)
		     end
		 else
		     evalTerm (e', push (stack, fn y => Fst y))

	   | Snd e' => 
		 if (isValue e') then
		     let val Pair (v1, v2) = e'
		     in
			 apply (v2, stack)
		     end
		 else
		     evalTerm (e', push (stack, fn y => Snd y))
		     
	   | Box e' => apply (Box e', stack)
		 
	   | Let (nil, e') => 
		 evalTerm (e', stack)

	   | Let ((NameDecl (x, tp))::ds, e2) => 
		 let val x' = Var.derived_var x
		     val s = Var.VarMap.insert (Var.VarMap.empty, x, Var x')
		     val (s', ds') = substDecls (s, ds)
		     val e2' = substTerm (s', e2)
		 in
		     evalTerm (Let (ds', e2'), stack)
		 end
		     
	   | Let ((ValDecl (x, e1))::ds, e2) => 
		 if (isValue e1) then 
		     let val s = Var.VarMap.insert (Var.VarMap.empty, x, e1)
			 val (s', ds') = substDecls (s, ds)
		     in
			 evalTerm (Let (ds', substTerm (s', e2)), stack)
		     end
		 else
		     evalTerm (e1, push (stack, fn y => Let ((ValDecl(x, y))::ds, e2)))

	   | Let ((BoxDecl (x, e1))::ds, e2) => 
		 if (isValue e1) then
		     let val (Box e') = e1
			 val s = Var.VarMap.insert (Var.VarMap.empty, x, e')
			 val (s', ds') = substDecls (s, ds)
		     in
			 evalTerm (Let (ds', substTerm (s', e2)), stack)
		     end
		 else
		     evalTerm (e1, push (stack, fn y => Let ((BoxDecl(x, y))::ds, e2)))

	   | Choose (Nu (x, e3)) => 
		 evalTerm (e3, stack)

	   | Choose e' => 
		 evalTerm (e', push (stack, fn y=>Choose y))

	   | Nu (x, e') => 
		 apply (Nu (x, e'), stack)

	   | Shift (x, k, e') => 
		 let val (l, c, stack') = pop stack
		 in
		     if Var.equal (x, l) then
			 let val y = Var.fromString_unique "x" 
			     val u = Var.fromString_unique "u"
			     val K = 
				 case c of
				     Return => Fn (y, Let ([BoxDecl (u, Var y)], Box (Var u)))
				   | Cont f => 
					 Fn (y,Let ([BoxDecl (u, Var y)], Box (f (Var u))))
			     val s = Var.VarMap.insert (Var.VarMap.empty, k, K)
			 in
			     evalTerm (substTerm (s, e'), stack')
			 end
		     else raise RuntimeError
		 end

	   | Reset (x, e') => 
		 let val stack' = ((x, Return)::stack)
		 in
		     evalTerm (e', stack')
		 end

	   | OfTp (e', _) => evalTerm (e', stack))




    fun evalDeclsPrint (decls, stack) = 
	(case decls of
	     nil => ()

	   | (NameDecl (x, tp)::ds) => 
		 let val _ = print ("name "^(Var.toString x)^"\n")
		 in
		     evalDeclsPrint (ds, stack)
		 end

	   | (ValDecl (x, e)::ds) => 
		 let val v = evalTerm (e, stack)
		     val _ = print ("val "^(Var.toString x)^" = "^(termToString v)^"\n")
		     val s = Var.VarMap.insert (Var.VarMap.empty, x, v)
		     val (s', ds') = substDecls (s, ds)
		 in
		     evalDeclsPrint (ds', stack)
		 end
	   
	   | (BoxDecl (x, e)::ds) => 
		 let val v = evalTerm (e, stack)
		 in
		     case v of
			 Box (e') => 
			     let val _ = print ("box "^(Var.toString x)^
						" = box ("^(termToString e')^")\n")
				 val s = Var.VarMap.insert (Var.VarMap.empty, x, e')
				 val (s', ds') = substDecls (s, ds)
			     in
				 evalDeclsPrint (ds', stack)
			     end
		       | _ => raise RuntimeError
		 end)

	     

    fun fromFile (file : string) = 
	let val ldecls = Parse.fromFile file
	in
	    let val _ = Typecheck.typechecks ldecls
		val decls = map toDecl ldecls
		val _ = evalDeclsPrint (decls, emptystack)
	    in
		()
	    end handle Error.Error msg =>  print msg
	end
	


end